#include <stdio.h>
#include <alloc.h>
#include <dos.h>
#include <conio.h>
#include <mem.h>
#include <process.h>
#include "palette.h"

typedef struct {
	char manufacturer;
	char version;
	char encoding;
	char bits_per_pixel;
	int xmin,ymin;
	int xmax,ymax;
	int hres;
	int vres;
	char palette[48];
	char reserved;
	char color_planes;
	int byte_per_line;
	int palette_type;
	char filler[58];
} PCXHEAD;

PCXHEAD header;
unsigned int width,depth;
unsigned int bytes,bits;
char palette[768];

void error(char *s);
void setvgapalette(char *p);
void unpackpcxfile(FILE *fp);
void readpcxline(char far *p, FILE *fp);
void viewpcx(char *name);
void video(char T);

void viewpcx(char *name)
{
  FILE *fp;

  if ((fp=fopen(name,"rb")) == NULL) error("Error opening the file.\n");

  if (fread((char *)&header, 1, sizeof(PCXHEAD), fp) != sizeof(PCXHEAD))
	error("Error reading the header.\n");

  if (header.manufacturer != 0x0a || header.version != 5)
	error("This is not a suitable PCX file.\n");

  if (header.bits_per_pixel == 1) bits=header.color_planes;
  else 				  bits=header.bits_per_pixel;

  if (bits != 8) error("This file has the wrong number of colors.\n");

  // find the palette
  if (!fseek(fp, -769L, SEEK_END)) {
	if (fgetc(fp) != 0x0c || fread(palette, 1, 768, fp) != 768)
	  error("Error reading palette.\n");
  } else error("Error seeking to palette.\n");

  fseek(fp, 128L, SEEK_SET);

  // allocate a big buffer
  width = (header.xmax-header.xmin)+1;
  depth = (header.ymax-header.ymin)+1;
  bytes = header.byte_per_line;

  unpackpcxfile(fp);
  fclose(fp);
}

void unpackpcxfile(FILE *fp)
{
	int i;

	// graphics on
	// screen mode 13H
	video(1);

	// set the palette
	setvgapalette(palette);

	// unpack the file directly to the VGA buffer
	for (i=0; i<depth; i++) {
	  // don't unpack more lines than there are on the screen
	  if (i >= 200) break;
	  readpcxline((char far *)MK_FP(0xA000, i*320), fp);
	}

	// graphics off
	// screen mode 3h
	video(0);
}

void readpcxline(char far *p, FILE *fp)
{
	int n=0, c,i;

	do {
	  // get a key byte
	  c=fgetc(fp) & 0xff;

	  // if it's a run of bytes field
	  if ((c & 0xc0) == 0xc0) {
		/* and off the high bits */
		i=c & 0x3f;
		/* get the run byte */
		c=fgetc(fp);
		/* run the byte */
		while(i--) p[n++]=c;
	  }
	  /* else just store it */
	  else p[n++]=c;

	} while(n < bytes);
}

void setvgapalette(char *p) // set the VGA palette to RGB buffer p
{
	int i;

	// convert eight bits to six bits
	for (i=0; i<768; i++) p[i] >>= 2;
	for (i=0; i<256; i++) setrgb(i, p[i*3], p[i*3+1], p[i*3+2]);
}

void error(char *s)
{
	puts(s);
	exit(1);
}

void video(char T)
{
  asm {
      mov     ah,12h
      mov     al,T
      mov     bl,36h
      int     10h
  }
}

